from dataclasses import dataclass
from typing import NamedTuple
from pathlib import PurePosixPath


class GCFContext:
    '''Just a fake type, to help auto-completion in IDE.
    The real type is 'google.cloud.functions_v1.context.Context', which only exists
    when the function is run inside Google Cloud Function runtime.
    '''
    # Ex: 83006d7a-af1b-4833-8bc4-ab4ac8a374f9-0
    event_id: str
    # Ex: 2019-12-14T10:18:29.911055Z
    timestamp: str
    # Ex: google.storage.object.finalize
    event_type: str
    resource: str


class ImgSize(NamedTuple):
    width: int
    height: int

    def __str__(self):
        return f'{self.width}x{self.height}'


@dataclass
class Thumbnail:
    content: bytes
    path: PurePosixPath
    size: ImgSize
    mimetype: str



import json
import time
from io import BytesIO
from pathlib import PurePosixPath
from typing import Dict
from datetime import datetime
from urllib.parse import quote_plus

import pendulum
import lazy_object_proxy
from logbook import Logger
from PIL import Image
from PIL import UnidentifiedImageError
from google.cloud import storage
from google.cloud import pubsub_v1
from google.cloud.storage import Bucket
from google.cloud.exceptions import NotFound
from google.cloud.pubsub_v1.publisher.futures import Future
from google.api_core.exceptions import GoogleAPICallError, RetryError

from . import __version__
from .common import GCFContext, ImgSize, Thumbnail
from .conf import get_monitored_paths, get_thumbnail_sizes, should_notify


THUMB_SUBFOLDER = 'thumbnails'
TOPIC_PREFIX = 'thumbnail-generated'
logger = Logger(__name__)
# Lazily instantiate Google Cloud client, so that it won't break unittest
store = lazy_object_proxy.Proxy(storage.Client)   # type: storage.Client


def build_thumbnail_path(original: PurePosixPath, size: ImgSize) -> PurePosixPath:
    '''
    Build file path for thumbnail from original file.

    Example: abc/photo.jpg -> abc/photo_512x512.jpg
    The "orignal" argument has PurePosixPath type because the path is in Google Cloud Storage context,
    not neccessary a real filesytem path.
    '''
    ext = original.suffix
    folder = original.parent
    return folder / THUMB_SUBFOLDER / f'{original.stem}_{size}{ext}'


def upload(bucket: storage.Bucket, thumb: Thumbnail) -> bool:
    blob = bucket.blob(str(thumb.path))
    blob.upload_from_string(thumb.content, thumb.mimetype)
    logger.info('Uploaded {}.', thumb.path)
    # TODO: Copy ACL from original image
    blob.make_public()
    meta = {'Generator': f'Thunagen v{__version__}'}
    blob.metadata = meta
    blob.update()
    logger.debug('Made {} public and set metadata {}', thumb.path, meta)
    return True


def create_thumbnail(orig: Image.Image, size: ImgSize, orpath: PurePosixPath) -> Thumbnail:
    img = orig.copy()
    mimetype = orig.get_format_mimetype()
    img.thumbnail(size)
    out = BytesIO()
    img.save(out, orig.format)
    out.seek(0)
    thumbpath = build_thumbnail_path(orpath, size)
    logger.debug('Thumbnail path: {}', thumbpath)
    return Thumbnail(out.getvalue(), thumbpath, size, mimetype)


def delete_thumbnails(bucket: Bucket, orpath: PurePosixPath):
    folder = orpath.parent
    prefix = folder / THUMB_SUBFOLDER / orpath.stem
    blobs = storage.list_blobs(bucket, prefix=str(prefix), fields='item(name)')
    bucket.delete_blobs(blobs, on_error=lambda b: logger.error('File {} seems to be deleted before.', b.name))


def notify_thumbnails_generated(project_id: str, original_path: str, generated: Dict[str, str]):
    publisher = pubsub_v1.PublisherClient()
    topic_path = publisher.topic_path(project_id, quote_plus(f'{TOPIC_PREFIX}/{original_path}'))
    logger.debug('To publish to: {}', topic_path)
    try:
        publisher.create_topic(topic_path)
    except (GoogleAPICallError, RetryError) as e:
        logger.error('Failed to create topic {}. Error: {}', topic_path, e)
        return
    data = json.dumps(generated).encode()
    future = publisher.publish(topic_path, data)  # type: Future
    # Google Cloud's Future object cannot be checked with Python concurent.futures module
    for i in range(4):
        if future.done():
            break
        time.sleep(1)


def is_thumbnail_missing_or_obsolete(thumb_path: PurePosixPath, orig_updated_time: datetime, bucket: Bucket) -> bool:
    thumb_blob = bucket.get_blob(str(thumb_path))
    if not thumb_blob:
        return True
    return orig_updated_time > (thumb_blob.updated or thumb_blob.time_created)


def generate_gs_thumbnail(data: dict, context: GCFContext):
    '''Background Cloud Function to be triggered by Cloud Storage'''
    # For fields of data, look in: https://cloud.google.com/storage/docs/json_api/v1/objects#resource
    event_type = context.event_type
    if event_type != 'google.storage.object.finalize' and event_type != 'google.storage.object.delete':
        # Not the event we want
        return
    filepath = data['name']   # type: str
    monitored_paths = get_monitored_paths()
    # If root folder "/" is in watchlist, accept all path
    if not any(filepath.startswith(p) for p in monitored_paths) and '/' not in monitored_paths:
        logger.debug('File {} is not watched. Ignore.', filepath)
        return
    filepath = PurePosixPath(filepath)
    folder_name = filepath.parent.name
    if folder_name == THUMB_SUBFOLDER:
        logger.info('The file {} is already a thumbnail. Ignore.', filepath)
        return
    content_type = data['contentType']  # type: str
    if not content_type.startswith('image/'):
        logger.debug('The file {} is not an image (content type {}). Ignore.', filepath, content_type)
        return
    bucket = store.get_bucket(data['bucket'])
    if event_type == 'google.storage.object.delete':
        delete_thumbnails(bucket, filepath)
        return
    try:
        blob = bucket.get_blob(str(filepath))
    except NotFound:
        logger.error('File {} was deleted by another job.', filepath)
        return
    filecontent = blob.download_as_string()
    try:
        file_last_uploaded = pendulum.parse(data['updated'] or data['timeCreated'])
    except (ValueError, TypeError):
        logger.debug('Event data missing time fields. {}', data)
        file_last_uploaded = pendulum.now()
    try:
        orig = Image.open(BytesIO(filecontent))    # type: Image.Image
    except UnidentifiedImageError:
        logger.error('This image {} is not supported by Pillow.', filepath)
        return
    generated = {}
    for size in get_thumbnail_sizes():   # type: ImgSize
        planned_thumbpath = build_thumbnail_path(filepath, size)
        if not is_thumbnail_missing_or_obsolete(planned_thumbpath, file_last_uploaded, bucket):
            logger.info('Thumbnail {} already exists and newer than reported uploaded time {}. '
                        'Perhap some other cloud function created it. Skip.', planned_thumbpath, file_last_uploaded)
            continue
        thumb = create_thumbnail(orig, size, filepath)
        success = upload(bucket, thumb)
        if success:
            generated[str(size)] = str(thumb.path)
    project = store.project
    logger.debug('Thumbnails generated: {}', generated)
    if should_notify() and generated:
        original_path = f'{bucket.name}/{blob.name}'
        notify_thumbnails_generated(project, original_path, generated)



import os
from typing import Tuple

from dotenv import load_dotenv, find_dotenv

from .common import ImgSize


load_dotenv(find_dotenv(usecwd=True))


def get_thumbnail_sizes() -> Tuple[ImgSize, ...]:
    var = os.getenv('THUMB_SIZES')
    if not var:
        return ()
    sizes = []
    for p in var.split(','):
        spec = p.strip()
        try:
            w, h = spec.split('x')
        except ValueError:
            continue
        try:
            sizes.append(ImgSize(int(w), int(h)))
        except ValueError:
            continue
    return tuple(sizes)


def get_monitored_paths() -> Tuple[str, ...]:
    var = os.getenv('MONITORED_PATHS')
    if not var:
        return ()
    configured_paths = tuple(p for p in var.split(':') if p)
    # If '/' (root) appears in list of paths, return only it.
    if '/' in configured_paths:
        return ('/',)
    return configured_paths


def should_notify() -> bool:
    return bool(os.getenv('NOTIFY_THUMBNAIL_GENERATED'))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['thunagen']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=7.0.0,<8.0.0',
 'google-cloud-pubsub>=1.1.0,<2.0.0',
 'google-cloud-storage>=1.24.1,<2.0.0',
 'importlib_metadata>=1.3.0,<2.0.0',
 'lazy_object_proxy>=1.4.3,<2.0.0',
 'logbook>=1.5.3,<2.0.0',
 'pendulum>=2.0.5,<3.0.0',
 'python-dotenv>=0.10.3,<0.11.0',
 'single-version>=1.1,<2.0']

setup_kwargs = {
    'name': 'thunagen',
    'version': '0.3.7',
    'description': 'Google Cloud function to generate thumbnail for images in Google Storage.',
    'long_description': '========\nThunagen\n========\n\n\n.. image:: https://madewithlove.now.sh/vn?heart=true&colorA=%23ffcd00&colorB=%23da251d\n.. image:: https://badge.fury.io/py/thunagen.svg\n   :target: https://pypi.org/project/thunagen/\n\n\nGoogle Cloud function to generate thumbnail for images in Google Storage.\n\nConvention\n----------\n\nThe thumbnails are placed in a folder "thumbnails" at the same place as original file.\n\nThe thumbnail size is appended to filename, right before the extention part. For example:\n\n\n.. code-block::\n\n    bucket/\n    └── folder/\n        ├── photo.jpg\n        └── thumbnails/\n            ├── photo_128x128.jpg\n            └── photo_512x512.jpg\n        ├── photo-missing-extension\n        └── thumbnails/\n            ├── photo-missing-extension_128x128\n            └── photo-missing-extension_512x512\n\n\nThe function expects these environment variables to be set:\n\n- ``THUMB_SIZES``: Size of thumbnails to be generated. Example: ``512x512,128x128``.\n\n- ``MONITORED_PATHS``: Folders (and theirs children) where the function will process the uploaded images. Muliple paths are separated by ":", like ``user-docs:user-profiles``. If you want to monitor all over the bucket, set it as ``/``.\n\n- ``NOTIFY_THUMBNAIL_GENERATED`` (optional): Tell Thunagen to notify after thumbnails are created.\n\nThe variables can be passed via *.env* file in the working directory.\n\nGet notified when thumbnails are generated\n------------------------------------------\n\nOther applications may want to be informed when the thumbnails are created. We support this by leveraging Google Cloud Pub/Sub service.\n\nAfter finishing generating thumbnail, if the ``NOTIFY_THUMBNAIL_GENERATED`` environment variable is set (with non-empty value), the function will publish a message to Pub/Sub. The message is sent to topic ``thumbnail-generated/{bucket_name}/{image_path}``, with the content being JSON string of thumbnail info (size and path). Example:\n\n- Topic: ``thumbnail-generated%2Fbucket%2Ffolder%2Fphoto.jpg`` (URL-encoded of "thumbnail-generated/bucket/folder/photo.jpg")\n\n- Message:\n\n    .. code-block:: json\n\n        {\n            "128x128": "folder/thumbnails/photo_128x128.jpg",\n            "512x512": "folder/thumbnails/photo_512x512.jpg"\n        }\n\nOther applications can subscribe to that topic to get notified. Google doesnot allow slash ("/") in topic name, so subscribed applications have to take care of URL-encode, decode the topic.\n\n\nWhy Thunagen\n------------\n\nI\'m aware that there is already a `Firebase extension <https://firebase.google.com/products/extensions/storage-resize-images>`_ for the same purpose.\nBut that extension, when doing its job, need to create a temporary file and in many cases, falling into race condition when the temporary file is deleted by another execution of the same cloud function. Thunagen, on the other hand, generates the file and uploads (back to Storage) on-the-fly (in memory), so it doesn\'t get into that issue.\n\n\nInstallation\n------------\n\nThunagen is distributed via PyPI. You can install it with ``pip``::\n\n    pip install thunagen\n\n\nInclude to your project\n-----------------------\n\nThunagen is provided without a *main.py* file, for you to incorporate more easily to your project, where you may have your own way to configure deployment environment (different bucket for "staging" and "production", for example).\n\nTo include Thunagen, from your *main.py*, do:\n\n.. code-block:: py\n\n    from thunagen.functions import generate_gs_thumbnail\n\n\nCredit\n------\n\nThunagen is brought to you by `Nguyễn Hồng Quân <https://github.com/hongquan>`_, from SunshineTech (Việt Nam).\n',
    'author': 'Nguyễn Hồng Quân',
    'author_email': 'ng.hong.quan@gmail.com',
    'maintainer': 'Nguyễn Hồng Quân',
    'maintainer_email': 'ng.hong.quan@gmail.com',
    'url': 'https://github.com/sunshine-tech/thunagen.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
